(** Unix file access permissions as a partially ordered domain. *)

open Logic

module Make =
  struct
    include Default

    let props () =
      {no_props "Permissions" with
       df = isok
     }

    type t = int (* like Unix modes *)

    let is_value _ = true

    let entails x y = x = y

    let le_min x y = (x land y) = x

    let le_max x y = (x land y) = x

    let le_overlap x y = x = y

    open Token

    let parse_rwx = parser
      | [<'Minus>] -> 0
      | [<'Ident s>] ->
	  match s with
	  | "rwx" -> 7
	  | "rw" -> 6
	  | "rx" -> 5
	  | "r" -> 4
	  | "wx" -> 3
	  | "w" -> 2
	  | "x" -> 1
	  | _ -> raise Stream.Failure

    let parse = parser
      | [<'Nat n>] -> n
      | [<u = parse_rwx; g = parse_rwx; o = parse_rwx>] -> (u lsl 6) lor (g lsl 3) lor o

    let print_rwx n =
      match n mod 8 with
      | 7 -> [Ident "rwx"]
      | 6 -> [Ident "rw"]
      | 5 -> [Ident "rx"]
      | 4 -> [Ident "r"]
      | 3 -> [Ident "wx"]
      | 2 -> [Ident "w"]
      | 1 -> [Ident "x"]
      | 0 -> [Minus]
      | _ -> assert false

    let print x =
      print_rwx (x lsr 6) @
      PP_tilda :: print_rwx (x lsr 3) @
      PP_tilda :: print_rwx x

  end
